import logging
import time
from vital_llm_cluster_mgr.runpod_client.runpod_client import RunpodClient
import threading
from datetime import datetime, timezone, timedelta
import requests

class VitalLLMClusterMgr:
    def __init__(self, *, runpod_api_key: str = None):
        self.runpod_api_key = runpod_api_key
        self.runpod_client = RunpodClient(self.runpod_api_key)

        self.running_pods = []
        self._lock = threading.Lock()

        self._stop_event = threading.Event()
        self._thread = threading.Thread(target=self._recurring_task, daemon=True)
        logging.info("starting recurring task thread...")
        self._thread.start()

    def get_running_pods(self):
        with self._lock:
            return self.running_pods

    def get_runpod_client(self):
        return self.runpod_client

    def _recurring_task(self):
        """A background thread method that runs every 5 minutes."""
        while not self._stop_event.is_set():
            self.do_recurring_work()
            # Wait for 5 minutes or until the stop event is set.
            self._stop_event.wait(300)  # 300 seconds = 5 minutes

    def shutdown_pod(self, pod):
        """Shut down the given pod using the adapted shutdown procedure."""
        pod_id = pod.get("id")
        if not pod_id:
            logging.info("Pod does not have an 'id'. Cannot shutdown.")
            return

        # Step 4: Stop the Pod
        logging.info("Stopping pod %s...", pod_id)
        self.runpod_client.stop_pod(pod_id)

        # Step 5: Wait 30 seconds before terminating
        logging.info("Waiting 30 seconds before terminating pod %s...", pod_id)
        time.sleep(30)

        # Step 6: Terminate the Pod
        logging.info("Terminating pod %s...", pod_id)
        self.runpod_client.terminate_pod(pod_id)

        # Step 7: Poll once (wait 5 seconds) to check if the pod is terminated.
        logging.info("Polling to confirm pod %s is terminated...", pod_id)
        time.sleep(5)
        running_pods = self.runpod_client.get_running_pods()
        pod_ids = {p["id"] for p in running_pods}
        if pod_id not in pod_ids:
            logging.info("Pod %s has been successfully terminated.", pod_id)
        else:
            logging.info("Pod %s is still present. Will try again in the next iteration.", pod_id)

    def do_recurring_work(self):
        """Retrieve running pods, check their /info endpoint,
        and shut them down if they've been inactive for over 30 minutes.
        """
        logging.info("Running recurring task")
        try:
            # Get the current list of running pods.
            running_pods = self.runpod_client.get_running_pods()
            with self._lock:
                self.running_pods = running_pods
            logging.info("Updated running pods: %s", running_pods)

            # Process each running pod.
            for pod in running_pods:
                logging.info("Processing running pod: %s", pod)
                pod_id = pod.get("id")
                if not pod_id:
                    continue

                # Build the pod's /info endpoint URL.
                # The URL format: https://{pod_id}-8000.proxy.runpod.net/info
                pod_url = f"https://{pod_id}-8000.proxy.runpod.net/info"
                try:
                    response = requests.get(pod_url, timeout=10)
                    if response.status_code == 200:
                        data = response.json()
                        last_time_used_str = data.get("last_time_used")

                        logging.info("Pod %s last used: %s", pod_id, last_time_used_str)

                        if last_time_used_str:
                            # Parse the ISO-formatted timestamp.
                            last_time_used = datetime.fromisoformat(last_time_used_str)
                            if last_time_used.tzinfo is None:
                                last_time_used = last_time_used.replace(tzinfo=timezone.utc)
                            now = datetime.now(timezone.utc)
                            # Compare the last time used with the current time.
                            if now - last_time_used > timedelta(minutes=30):
                                logging.info("Pod %s last used over 30 minutes ago. Shutting it down.", pod_id)
                                self.shutdown_pod(pod)
                        else:
                            logging.info("Pod %s did not return a 'last_time_used' value.", pod_id)
                    else:
                        logging.info("Failed to fetch /info for pod %s: HTTP %s (Potentially starting or shutting down)", pod_id, response.status_code)
                except Exception as e:
                    logging.info("Error contacting pod %s /info endpoint: %s", pod_id, e)
        except Exception as e:
            logging.info("Error during recurring task: %s", e)

    def stop_recurring_task(self):
        """Call this method to stop the recurring task gracefully."""
        self._stop_event.set()
        self._thread.join()

    # run periodic task to get current runpod servers, grouped by type

    # functions to return the cached list of servers of a particular type

    # rest resource service to use this lookup to select a random one from the list

    # longer term, we may have state on the vllm server that we want to be sticky
    # so we could help route requests from a customer-id (account-id) to a particular instance
    # also can be more clever with routing to route to the least busy server

    # once deployed in serverless they that should handle routing, but wouldnt handle the sticky session part

    # have method that keeps track of last time an instance used

    # task shuts down instances that have not been used in 30 minutes
    # turn this off once in production

    # add call that warns up an instance of a given type/model/size


